# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018-19  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload data to WeatherCloud.

* Create account: https://weathercloud.net/
* Additional dependency: http://docs.python-requests.org/
* Example ``weather.ini`` configuration::

    [weathercloud]
    deviceid = XXXXXXXXXXXX
    devicekey = XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    internal = True

    [logged]
    services = ['weathercloud', 'metoffice']

.. _WeatherCloud: http://www.weathercloud.net/

"""

from __future__ import absolute_import, unicode_literals

from ast import literal_eval
from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import sys
if sys.version_info[0] < 3:
    from httplib import responses
else:
    from http.client import responses

import requests

from pywws.conversions import usaheatindex, wind_mph
import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.LiveDataService):
    config = {
        'deviceid' : ('',      True, 'wid'),
        'devicekey': ('',      True, 'key'),
        'internal' : ('False', True, None),
        }
    fixed_data = {'ver': pywws.__version__, 'type': '481'}
    interval = timedelta(seconds=600)
    logger = logger
    service_name = service_name
    template = """
#live#
#temp_out
    "'temp'     : '%.0f'," "" "scale(x, 10.0)"#
#calc "wind_chill(data['temp_out'], data['wind_ave'])"
    "'chill'    : '%.0f'," "" "scale(x, 10.0)"#
#calc "dew_point(data['temp_out'], data['hum_out'])"
    "'dew'      : '%.0f'," "" "scale(x, 10.0)"#
#calc "usaheatindex(data['temp_out'], data['hum_out'])"
    "'heat'     : '%.0f'," "" "scale(x, 10.0)"#
#hum_out
    "'hum'      : '%.d',"#
#wind_ave
    "'wspdavg'  : '%.0f'," "" "scale(x, 10.0)"#
#wind_ave
    "'wspd'     : '%.0f'," "" "scale(x, 10.0)"#
#wind_gust
    "'wspdhi'   : '%.0f'," "" "scale(x, 10.0)"#
#wind_dir
    "'wdiravg'  : '%.0f'," "" "winddir_degrees(x)"#
#wind_dir
    "'wdir'     : '%.0f'," "" "winddir_degrees(x)"#
#rel_pressure
    "'bar'      : '%.0f'," "" "scale(x, 10.0)"#
#calc "rain_day(data)"
    "'rain'     : '%.0f'," "" "scale(x, 10.0)"#
#calc "rain_hour(data)"
    "'rainrate' : '%.0f'," "" "scale(x, 10.0)"#
#idx
    "'time'     : '%Y%m%d %H%M%S',"#
"""

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # extend template
        if context.params.get('config', 'ws type') == '3080':
            self.template += """
#illuminance
    "'solarrad': '%.0f'," "" "scale(illuminance_wm2(x), 10.0)"#
#uv
    "'uvi'     : '%.0f'," "" "scale(x, 10.0)"#
"""
        if literal_eval(self.params['internal']):
            self.template += """
#temp_in
    "'tempin'  : '%.0f'," "" "scale(x, 10.0)"#
#hum_in
    "'humin'   : '%.d',"#
#calc "dew_point(data['temp_in'], data['hum_in'])"
    "'dewin'   : '%.0f'," "" "scale(x, 10.0)"#
#calc "usaheatindex(data['temp_in'], data['hum_in'])"
    "'heatin'  : '%.0f'," "" "scale(x, 10.0)"#
"""

    @contextmanager
    def session(self):
        with requests.Session() as session:
            yield session

    def valid_data(self, data):
        return any([data[x] is not None for x in (
            'wind_dir', 'wind_ave', 'wind_gust', 'hum_out', 'temp_out',
            'temp_in', 'hum_in', 'rel_pressure')])

    def upload_data(self, session, prepared_data={}):
        url = 'http://api.weathercloud.net/v01/set'
        try:
            rsp = session.get(url, params=prepared_data, timeout=60)
        except Exception as ex:
            return False, repr(ex)
        text = rsp.text.strip()
        if text == '200':
            return True, 'OK'
        if text in responses:
            return False, '{} ({})'.format(responses[text], text)
        return False, 'unknown error ({})'.format(text)


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Post messages to Mastodon.

Mastodon_ is a micro-blogging system that at first sight looks a bit
like Twitter. In many ways it's quite different though. This module
sends "toots", with optional image files, typically to report on weather
conditions every hour.

* Create account: https://joinmastodon.org/
* Additional dependency: https://mastodonpy.readthedocs.io/
* Example ``weather.ini`` configuration::

    [mastodon]
    handle = kt19weather@botsin.space
    client_id = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    client_secret = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    access_token = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    [hourly]
    text = ['toot.txt', 'tweet_media.txt', '24hrs.txt']
    plot = ['tweet.png.xml', '24hrs.png.xml', 'rose_12hrs.png.xml']
    services = [('mastodon', 'toot.txt'), ('twitter', 'tweet_media.txt'),
                ('ftp', '24hrs.png', 'rose_12hrs.png', '24hrs.txt')]

Create an account
-----------------

Before creating a Mastodon account for your weather reports you need to
choose an "instance". Mastodon is a federated system, running on many
interconnected servers, each of which is called an instance. Choose one
whose rules allow "bots" (i.e. automated posts) such as `botsin.space`_.
After creating an account, edit its profile and make sure the "this is a
bot account" box is selected.

The :py:mod:`pywws.service.mastodon` module requires you to install an
additional dependency::

    sudo pip install mastodon.py

Authorise pywws
---------------

Before you can send "toots" you need to authorise pywws to post to your
account. If you run pywws on a low power device such as a Raspberry Pi,
you may find it easier to run this authorisation step on another
computer, as long as it has the required dependencies installed. You can
use an empty 'data' directory -- a ``weather.ini`` file will be created
whose contents can be copied into your real ``weather.ini`` file using
any text editor.

Make sure no other pywws software is running, then run the module with
the ``-r`` option. (Remember to replace ``data_dir`` with your data
directory.) ::

    python -m pywws.service.mastodon -r data_dir

The first time you do this it will probably crash because you haven't
set your Mastodon "handle" in ``weather.ini``. Edit ``weather.ini`` and
add your handle as shown above, then run the module with the ``-r``
option again.

This will open a web browser window (or give you a URL to copy to your
web browser) where you can log in to your Mastodon account and authorise
pywws to post. If the login is successful the browser will display a
long code string which you then copy to pywws::

    jim@brains:~$ python3 -m pywws.service.mastodon -r weather_data
    07:45:34:pywws.logger:pywws version 18.8.0, build 1564 (5bfc528)
    Please enter the auth code shown in your web browser: 12573ba24341b5de2a1f2930fc93889c3576af7b50ecd9d713fa502d773805b4
    jim@brains:~$ 

The ``access_token`` value stored in ``weather.ini`` gives access to
your Mastodon account and should be kept confidential.

Create a template
-----------------

A "toot" is a short text of up to 500 characters. It's up to you what to
put in your "toots" but an example is included to get your started. Copy
the example template ``toot.txt`` to your template directory, then edit
it to suit your preferences. (At least change the hashtags to suit your
location.) You should also check it uses the same text encoding as your
other templates. The example template includes a ``media`` line to send
a graph. Either remove this or copy the example graph template
``tweet.png.xml`` to your graph templates directory, if you don't
already have one there.

Now generate a toot file from your template, for example::

    python -m pywws.template ~/weather/data ~/weather/templates/toot.txt toot.txt
    cat toot.txt

Post your first toot
--------------------

Now you are ready to run :py:mod:`pywws.service.mastodon`::

    python -m pywws.service.mastodon ~/weather/data toot.txt

If this works, your new Mastodon account will have posted its first
weather report. (You can delete the toot.txt file now.)

Add Mastodon posts to your hourly tasks
---------------------------------------

Edit the ``[hourly]`` section in ``weather.ini``. If your toots include
one or more graphs you need to add the graph templates to the ``plot``
list. Note that if you reuse your Twitter graph you only need to
generate it once. Add your toot template to the ``text`` list. Finally,
add ``mastodon`` to the ``services`` list, with an option specifying the
template processing result. For example::

    [hourly]
    text = ['toot.txt']
    plot = ['tweet.png.xml']
    services = [('mastodon', 'toot.txt')]

You could use the ``[logged]``, ``[12 hourly]`` or ``[daily]`` sections
instead, but I think ``[hourly]`` is most appropriate for Mastodon
updates.

Add more images
---------------

Mastodon allows up to four images per "toot", so you could add more
graphs, or a webcam image, or a weather forecast icon. Use one ``media``
line per image at the start of your toot template. You need to give the
full path of any files that are not in your "output" directory (a
subdirectory of your work directory called ``output``).

.. _botsin.space: https://botsin.space/
.. _Mastodon: https://joinmastodon.org/

"""

from __future__ import absolute_import, print_function

import codecs
from contextlib import contextmanager
import logging
import os
import sys

from mastodon import Mastodon

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.FileService):
    config = {
        'handle'      : ('',   True, None),
        'access_token': (None, True, None),
        }
    logger = logger
    service_name = service_name

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # get default character encoding of template output
        self.encoding = context.params.get(
            'config', 'template encoding', 'iso-8859-1')

    @contextmanager
    def session(self):
        yield Mastodon(access_token=self.params['access_token'],
                       api_base_url=self.params['handle'].split('@')[-1])

    def upload_file(self, session, filename):
        with codecs.open(filename, 'r', encoding=self.encoding) as toot_file:
            toot = toot_file.read()
        media = []
        while toot.startswith('media'):
            media_item, toot = toot.split('\n', 1)
            media_item = media_item.split()[1]
            if not os.path.isabs(media_item):
                media_item = os.path.join(self.context.output_dir, media_item)
            media.append(media_item)
        media_ids = []
        try:
            for media_item in media:
                rsp = session.media_post(media_item)
                media_ids.append(rsp['id'])
            rsp = session.status_post(status=toot, media_ids=media_ids)
        except Exception as ex:
            return False, repr(ex)
        return True, 'OK'

    def register(self):
        import webbrowser

        self.check_params('handle')
        api_base_url = self.params['handle'].split('@')[-1]
        # get client data
        client_id = self.context.params.get(service_name, 'client_id')
        client_secret = self.context.params.get(service_name, 'client_secret')
        if (not client_id) or (not client_secret):
            client_id, client_secret = Mastodon.create_app(
                'pywws', scopes=['write'], api_base_url=api_base_url)
            self.context.params.set(service_name, 'client_id', client_id)
            self.context.params.set(service_name, 'client_secret', client_secret)
        # create api
        api = Mastodon(client_id=client_id, client_secret=client_secret,
                       api_base_url=api_base_url)
        # authorise
        auth_request_url = api.auth_request_url(scopes=['write'])
        if not webbrowser.open(auth_request_url, new=2, autoraise=0):
            print('Please use a web browser to open the following URL')
            print(auth_request_url)
        if sys.version_info[0] >= 3:
            input_ = input
        else:
            input_ = raw_input
        code = input_('Please enter the auth code shown in your web browser: ')
        code = code.strip()
        # log in
        access_token = api.log_in(code=code, scopes=['write'])
        self.context.params.set(service_name, 'access_token', access_token)


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Store weather data in an SQLite3 database file.

Introduction
------------

This module is an alternative to the pywws file based storage system.
It stores data using the mature SQLite3 database system, but with
underlying queries wrapped so the user need not know about them.

It should also be possible for this module to form the basis of a
full client-server based SQL module using, for example, MySQL etc.

The Python builtin sqlite3 module is used which has a threadsafety of 1,
therefore this module creates a connection with every Store (sub)class
instance. This however brings concurrancy issues and so this module makes
use of the underlying sqlite3's Write-Ahead-Loging and Shared Cache modes
to relieve this. These rely on up to date sqlite3 libraries and may not
work on older networked drives which do not support the right locking
semantics required by sqlite3.


The external API is as per the original pywws file store, but with
some enhancements so as to behave more like a mapping container (dict).

From a "user" point of view, the data is accessed as a cross between a
list and a dictionary. Each data record is indexed by a
:py:class:`datetime.datetime` object (dictionary behaviour), but
records are stored in order and can be accessed as slices (list
behaviour).

For example, to access the hourly data for Christmas day 2009, one
might do the following::

  from datetime import datetime
  import pywws.filedata
  hourly = pywws.filedata.HourlyStore("weather_data")
  for data in hourly[datetime(2009, 12, 25):datetime(2009, 12, 26)]:
    print(data["idx"], data["temp_out"])

Some more examples of data access::

  # get value nearest 9:30 on Christmas day 2008
  data[data.nearest(datetime(2008, 12, 25, 9, 30))]
  # get entire array, equivalent to data[:]
  data[datetime.min:datetime.max]
  # get last 12 hours worth of data
  data[datetime.utcnow() - timedelta(hours=12):]

Note that the :py:class:`datetime.datetime` index is in UTC. You may
need to apply an offset to convert to local time.

The module provides five classes to store different data.
:py:class:`RawStore` takes "raw" data from the weather station;
:py:class:`CalibStore`, :py:class:`HourlyStore`, :py:class:`DailyStore`
and :py:class:`MonthlyStore` store processed data (see
:py:mod:`pywws.process`). All are derived from the same ``CoreStore``
class, they only differ in the keys and types of data stored in each
record.

Detailed API
------------

"""

import sqlite3
import os.path
from threading import RLock
from datetime import date, datetime, timedelta

import pytz

from pywws.weatherstation import WSDateTime, WSFloat, WSInt, WSStatus

# Data type adapt: Python ==> SQLite3
def _adapt_WSDateTime(dt):
    """Return unix timestamp of the datetime like input.
    If conversion overflows high, return sint64_max ,
    if underflows, return 0
    """
    try:
        ts = int(
            (dt.replace(tzinfo=pytz.utc)
            - datetime(1970,1,1,tzinfo=pytz.utc)
            ).total_seconds()
        )
    except (OverflowError,OSError):
        if dt < datetime.now():
            ts = 0
        else:
            ts = 2**63-1
    return ts

def _adapt_WSStatus(status):
    """Return integer represending WSStatus dictionary input"""
    return int(WSStatus.to_csv(status))

# Data type convert SQLite3 ==> Python
def _convert_WSDateTime(b):
    """Return a WSDateTime object for a given unix timestamp"""
    return WSDateTime.utcfromtimestamp(int(b))

def _convert_WSStatus(b):
    """Return a WSStatus dictionary for a given number"""
    return WSStatus.from_csv(b)

def _convert_WSFloat(b):
    """Return a WSFloat for the given input"""
    return WSFloat(b)

def _convert_WSInt(b):
    """Return WSInt for the given input"""
    return WSInt(b)

sqlite3.enable_shared_cache(True)

sqlite3.register_adapter(datetime, _adapt_WSDateTime)
sqlite3.register_adapter(WSDateTime, _adapt_WSDateTime)
sqlite3.register_adapter(WSStatus, _adapt_WSStatus)
# Ensures SQLite handles these special types and treats them as standard float
sqlite3.register_adapter(WSFloat, float)
# Wnsures SQLite handles these special types and treats them as standard int
sqlite3.register_adapter(WSInt, int)

sqlite3.register_converter("WSDateTime", _convert_WSDateTime)
sqlite3.register_converter("WSStatus", _convert_WSStatus)
sqlite3.register_converter("WSFloat", _convert_WSFloat)
sqlite3.register_converter("WSInt", _convert_WSInt)


class CoreStore(object):
    """Provides a dictionary/list like interface
    to an underlying SQLite3 database
    """
    # Overrides type conversion based on incorrect sql type
    # (i.e. for the primary key)
    conv = {"idx":None}
    key_list = tuple(conv.keys())
    table = ""
    _keycol = "idx"
    if len(conv) == 0:
        raise KeyError("No coloumns are defined.")
    if _keycol not in key_list:
        # Check that the key coloumn is present
        raise KeyError(
            "Key coloumn '{}' is not in the key list".format(keycol)
        )

    def __init__(self, dir_name):
        key_list = self.key_list
        conv = self.conv
        table = self.table
        keycol = self._keycol
        dbpath = os.path.abspath(os.path.join(dir_name, "pywws.db"))
        self._connection = sqlite3.connect(
            dbpath,
            detect_types=sqlite3.PARSE_COLNAMES
        )
        self._connection.row_factory = sqlite3.Row
        con = self._connection

        if con.execute(
            "PRAGMA journal_mode=WAL"
        ).fetchone()["journal_mode"] != "wal":
            raise TypeError("Database is not in Write-Ahead-Log mode")
        # Create the table if its not already there.
        # Assume all data fields are NUM so that SQLite can optimise storage
        # as it will choose the smallest integer representation between 8-64bit,
        # while floats are always 64bit. Set idx as a unique integer primary
        # key so searches are faster, storage requirements smaller,
        # the rowid coloumn is eliminated so there is no need for secondary
        # indices. Suitable converters/adapters are then applied.
        if con.execute(
            """SELECT COUNT(*) FROM sqlite_master
            WHERE type="table" AND name=?;""",
            (table,)
        ).fetchone()[0] == 0:
            con.executescript(
                """CREATE TABLE IF NOT EXISTS {table} (
                {keycol} INTEGER PRIMARY KEY, {coloumns} );""".format(
                    table=table,
                    keycol=keycol,
                    coloumns=", ".join(
                        "{} NUM".format(key)
                        for key in conv
                        if key != keycol
                    )
                )
            )

        # Get all coloumns from the database
        sql_key_list = tuple(
            (row["name"],row["pk"])
            for row in con.execute(
                "SELECT name, pk FROM PRAGMA_TABLE_INFO(?);",
                (table,)
            )
        )

        # Fetch the primary key - check there is only one
        sql_pk = tuple(key[0] for key in sql_key_list if key[1] == 1)
        if len(sql_pk) != 1 or sql_pk[0] != keycol:
            raise KeyError(
                "Mismatch between database primary key and what was expected"
            )

        # Convert this to just a set of keys
        sql_key_list = set(key[0] for key in sql_key_list)
        # Check that no coloumns are missing
        if not set(conv.keys()) <= sql_key_list:
            raise KeyError(
                "Mismatch between database coloumns and what was expected"
            )

        # SQL snippet which casts all coloumns to the correct data types
        # for SELECT * type queries
        self.selallcols = ", ".join(
            '{col} AS "{col} [{conv}]"'.format(
                col=col,
                conv=conv[col]
            ) for col in key_list
        )
        # SQL snippet which casts the key coloum to the correct data type
        # for SELECT {keycol} type queries
        self.selkeycol = '{keycol} AS "{keycol} [{conv}]"'.format(
            keycol=keycol,
            conv=conv[keycol]
        )

    def __del__(self):
        """Prior to object being deleted, update SQLite statistics,
        then close connection gracefully
        """
        with self._connection as con:
            con.executescript(
                """ANALYZE {}; PRAGMA wal_checkpoint(TRUNCATE);
                """.format(self.table)
            )
        self._connection.close()

    def __len__(self):
        """Return the exact number of records in the table"""
        # Direct count of all records - could be slow.
        return self._connection.execute(
            "SELECT COUNT(*) FROM {};".format(self.table)
        ).fetchone()[0]
    
    def __length_hint__(self):
        """Return the approximate table size based on internal database
        statistics if present, otherwise, find the actual length
        """
        # Assuming the database has been analyzed before, the stat1 table
        # should contain the total row count for the table as the first number
        # in the stat coloumn from when it was last analyzed. Very fast but may
        # not be up to date
        try:
            return int(self._connection.execute(
                "SELECT stat FROM sqlite_stat1 WHERE tbl=?;",
                (self.table,)
                ).fetchone()[0].split(' ')[0]
            )
        except (TypeError,sqlite3.OperationalError):
            return self.__len__()

    def _predicate(self, i):
        """Given a valid datetime or slace, return the predicate portion
        of the SQL query, a boolean indicating whether multiple items are
        expected from the result, and a dictionary of parameters for the query
        """
        if isinstance(i, slice):
            if i.step is not None:
                raise TypeError("Slice step not permitted")
            if ( (i.start is not None and not isinstance(i.start, datetime))
                or (i.stop is not None and not isinstance(i.stop, datetime))
            ):
                raise TypeError(
                    "Slice indices must be {} or None".format(datetime)
                )
            if i.start is not None and i.stop is not None:
                if i.start > i.stop:
                    raise ValueError(
                        "Start index is greater than the End index"
                    )
                else:
                    # Substitution of the key coloumn, but the
                    # parameters themselves will be substituted by sqlite3
                    predicate = "WHERE {} BETWEEN :start AND :stop".format(
                        self._keycol
                    )
            elif i.start is not None:
                # i.stop will also be None
                predicate = "WHERE {} >= :start".format(self._keycol)
            elif i.stop is not None:
                # i.start will also be None
                predicate = "WHERE {} <= :stop".format(self._keycol)
            else:
                # both are None, so equivelent to wanting everything
                predicate = ""
            multi = True
            pred = {"start": i.start, "stop": i.stop}
        elif isinstance(i, datetime):
            # Substitution of the key coloumn, but the
            # parameters themselves will be substituted by sqlite3
            predicate = "WHERE {} = :key".format(self._keycol)
            multi = False
            pred = {"key": i}
        else:
            # not a slice or a datetime object
            raise TypeError("List indices must be {}".format(datetime))
        # predicate is the end of the query string.
        # multi is a boolean indicating whether the result should be iterable
        # or not. pred is a dict of the parameters for substitution
        return (predicate, multi, pred)

    def __getitem__(self, i):
        """Return the data item or items with index i.
        
        i must be a datetime object or a slice. If i is a single datetime
        then a value with that index must exist.
        """
        predicate, multi, params = self._predicate(i)
        results = self._connection.execute(
            "SELECT {} FROM {} {};".format(
                self.selallcols,self.table,predicate), params)
        if multi:#
            # If multiple items are expected, give a generator
            return (dict(row) for row in results)
        else:
            # If one item is expected, return it directly
            item = results.fetchone()
            return self.__missing__(i) if item is None else dict(item)

    def __contains__(self, i):
        """Return True if i is in the table, else False"""
        if self._connection.execute(
            "SELECT count(*) FROM {} WHERE {} = ?;".format(
                self.table,
                self._keycol
            ),(i,)).fetchone()[0] > 0:
            return True
        else:
            return False

    def __missing__(self, i):
        raise IndexError(i)

    def __setitem__(self, i, x):
        """Store a value x with index i.
        
        i must be a datetime object. If there is already a value with
        index i, it is overwritten.
        """
        if not isinstance(i, datetime):
            raise TypeError("'{}' is not a datetime object".format(i))
        elif not isinstance(x, dict):
            raise TypeError("Values not a dictionary")
        x[self._keycol] = i
        with self._connection as con:
            con.execute(
                """INSERT OR REPLACE INTO {table} ({keylist})
                VALUES (:{vallist});
                """.format(
                    table=self.table,
                    keylist=", ".join(x),
                    vallist=", :".join(x)
                ), x)

    def update(self, i):
        """D.update(E) -> None.  Update D from iterable E with pre-existing
        items being overwritten.
        
        Elements in E are assumed to be dicts containing the primary key to
        allow the equivelent of:
        for k in E: D[k.primary_key] = k
        """
        key_list = self.key_list
        keynone = {key:None for key in key_list}
        # Generator which fills in missing data from the original iterator
        def datagen(i):
            for datum in i:
                tmp = keynone.copy()
                tmp.update(datum)
                yield tmp
        with self._connection as con:
            con.executemany(
                """INSERT OR REPLACE INTO {table} ({keylist})
                VALUES (:{vallist});
                """.format(table=self.table,
                    keylist=", ".join(self.key_list),
                    vallist=", :".join(self.key_list)
                ), datagen(i))

    def __delitem__(self, i):
        """Delete the data item or items with index i.
        
        i must be a datetime object or a slice. If i is a single datetime
        then a value with that index must exist.
        """
        predicate, multi, params = self._predicate(i)
        with self._connection as con:
            if con.execute("DELETE FROM {} {};".format(
                self.table,
                predicate
            ), params).rowcount == 0 and multi is False:
                raise KeyError(i)

    def before(self, i):
        """Return datetime of newest existing data record whose datetime
        is < idx. If no such record exists, return None.
        """
        if not isinstance(i, datetime):
            raise TypeError("'{}' is not a datetime object".format(i))
        else:
            result = self._connection.execute(
                """SELECT {selkeycol} FROM {table} WHERE
                {keycol} < :key ORDER BY {keycol} DESC LIMIT 1;
                """.format(
                    selkeycol=self.selkeycol,
                    table=self.table,
                    keycol=self._keycol
                ), {"key":i}).fetchone()
        return result[self._keycol] if result is not None else None

    def after(self, i):
        """Return datetime of oldest existing data record whose datetime
        is >= idx. If no such record exists, return None.
        """
        if not isinstance(i, datetime):
            raise TypeError("'{}' is not a datetime object".format(i))
        else:
            result = self._connection.execute(
                """SELECT {selkeycol} FROM {table} WHERE
                {keycol} >= :key ORDER BY {keycol} ASC LIMIT 1;""".format(
                    selkeycol=self.selkeycol,
                    table=self.table,
                    keycol=self._keycol
                ), {"key":i}).fetchone()
        return result[self._keycol] if result is not None else None

    def nearest(self, i):
        """Return datetime of record whose datetime is nearest idx."""
        if not isinstance(i, datetime):
            raise TypeError("'{}' is not a datetime object".format(i))
        else:
            result = self._connection.execute(
                """SELECT {selkeycol} FROM {table} ORDER BY
                ABS({keycol}-:key) ASC LIMIT 1;""".format(
                    selkeycol=self.selkeycol,
                    table=self.table,
                    keycol=self._keycol
                ), {"key":i}).fetchone()
        return result[self._keycol] if result is not None else None

    def flush(self):
        """Commits any uncommitted data and then flushes the write ahead log"""
        with self._connection as con:
            # execute script has implied commit before it
            con.executescript("PRAGMA wal_checkpoint(TRUNCATE);")

    def keys(self):
        """D.keys() -> a set-like object providing a view on D's keys"""
        return set(
            row[self._keycol] for row in self._connection.execute(
                """SELECT DISTINCT {} FROM {} ORDER BY {} ASC;""".format(
                    self.selkeycol,
                    self.table,
                    self._keycol
                )
            )
        )

    def __iter__(self):
        """Iterates over all rows in ascending order of key coloumn"""
        for row in self._connection.execute(
            """SELECT {} FROM {} ORDER BY {} ASC;""".format(
                self.selallcols,
                self.table,
                self._keycol
            )
        ):
            yield dict(row)

    def __reversed__(self):
        """Iterates over all rows in decending order of key coloumn"""
        for row in self._connection.execute(
            """SELECT {} FROM {} ORDER BY {} DESC;""".format(
                self.selallcols,
                self.table,
                self._keycol
            )
        ):
            yield dict(row)

    def values(self):
        """D.values() -> an object providing a view on D's values"""
        for item in self.__iter__():
            yield item

    def items(self):
        """D.items() -> a set-like object providing a view on D's items"""
        keycol = self._keycol
        for row in self.__iter__():
            yield (row[keycol], dict(row))

    def get(self, key, default=None):
        """D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None."""
        try:
            return self[key]
        except (KeyError, IndexError):
            return default

    def clear(self):
        """S.clear() -> None -- remove all items from S"""
        with self._connection as con:
            con.execute("DELETE FROM {};".format(self.table))

    def setdefault(self, key, default=None):
        """D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D"""
        try:
            return self[key]
        except (KeyError, IndexError):
            self[key] = default
        return default

    __marker = object()
    def pop(self, key, default=__marker):
        """D.pop(k[,d]) -> v
        
        Remove specified key and return the corresponding value.
        If key is not found, d is returned if given,
        otherwise KeyError is raised.
        """
        try:
            value = self[key]
            del self[key]
            return value
        except (KeyError, IndexError):
            if default is self.__marker:
                raise
            else:
                return default

    def popitem(self):
        """D.popitem() -> (k, v)
        
        Remove and return some (key, value) pair
        as a 2-tuple; but raise KeyError if D is empty.
        """
        try:
            value = next(iter(self))
            key = value[self._keycol]
        except StopIteration:
            raise KeyError
        del self[key]
        return key, value


class RawStore(CoreStore):
    """Stores raw weather station data."""
    table = "raw"
    conv = {
        "idx"          : "WSDateTime",
        "delay"        : "WSInt",
        "hum_in"       : "WSInt",
        "temp_in"      : "WSFloat",
        "hum_out"      : "WSInt",
        "temp_out"     : "WSFloat",
        "abs_pressure" : "WSFloat",
        "wind_ave"     : "WSFloat",
        "wind_gust"    : "WSFloat",
        "wind_dir"     : "WSInt",
        "rain"         : "WSFloat",
        "status"       : "WSStatus",
        "illuminance"  : "WSFloat",
        "uv"           : "WSInt"
    }
    key_list = tuple(conv.keys())


class CalibStore(CoreStore):
    """Stores "calibrated" weather station data."""
    table = "calib"
    conv = {
        "idx"          : "WSDateTime",
        "delay"        : "WSInt",
        "hum_in"       : "WSInt",
        "temp_in"      : "WSFloat",
        "hum_out"      : "WSInt",
        "temp_out"     : "WSFloat",
        "abs_pressure" : "WSFloat",
        "rel_pressure" : "WSFloat",
        "wind_ave"     : "WSFloat",
        "wind_gust"    : "WSFloat",
        "wind_dir"     : "WSFloat",
        "rain"         : "WSFloat",
        "status"       : "WSStatus",
        "illuminance"  : "WSFloat",
        "uv"           : "WSInt"
    }
    key_list = tuple(conv.keys())


class HourlyStore(CoreStore):
    """Stores hourly summary weather station data."""
    table = "hourly"
    conv = {
        "idx"            : "WSDateTime",
        "hum_in"         : "WSInt",
        "temp_in"        : "WSFloat",
        "hum_out"        : "WSInt",
        "temp_out"       : "WSFloat",
        "abs_pressure"   : "WSFloat",
        "rel_pressure"   : "WSFloat",
        "pressure_trend" : "WSFloat",
        "wind_ave"       : "WSFloat",
        "wind_gust"      : "WSFloat",
        "wind_dir"       : "WSFloat",
        "rain"           : "WSFloat",
        "illuminance"    : "WSFloat",
        "uv"             : "WSInt"
    }
    key_list = tuple(conv.keys())


class DailyStore(CoreStore):
    """Stores daily summary weather station data."""
    table = "daily"
    conv = {
        "idx"                : "WSDateTime",
        "start"              : "WSDateTime",
        "hum_out_ave"        : "WSFloat",
        "hum_out_min"        : "WSInt",
        "hum_out_min_t"      : "WSDateTime",
        "hum_out_max"        : "WSInt",
        "hum_out_max_t"      : "WSDateTime",
        "temp_out_ave"       : "WSFloat",
        "temp_out_min"       : "WSFloat",
        "temp_out_min_t"     : "WSDateTime",
        "temp_out_max"       : "WSFloat",
        "temp_out_max_t"     : "WSDateTime",
        "hum_in_ave"         : "WSFloat",
        "hum_in_min"         : "WSInt",
        "hum_in_min_t"       : "WSDateTime",
        "hum_in_max"         : "WSInt",
        "hum_in_max_t"       : "WSDateTime",
        "temp_in_ave"        : "WSFloat",
        "temp_in_min"        : "WSFloat",
        "temp_in_min_t"      : "WSDateTime",
        "temp_in_max"        : "WSFloat",
        "temp_in_max_t"      : "WSDateTime",
        "abs_pressure_ave"   : "WSFloat",
        "abs_pressure_min"   : "WSFloat",
        "abs_pressure_min_t" : "WSDateTime",
        "abs_pressure_max"   : "WSFloat",
        "abs_pressure_max_t" : "WSDateTime",
        "rel_pressure_ave"   : "WSFloat",
        "rel_pressure_min"   : "WSFloat",
        "rel_pressure_min_t" : "WSDateTime",
        "rel_pressure_max"   : "WSFloat",
        "rel_pressure_max_t" : "WSDateTime",
        "wind_ave"           : "WSFloat",
        "wind_gust"          : "WSFloat",
        "wind_gust_t"        : "WSDateTime",
        "wind_dir"           : "WSFloat",
        "rain"               : "WSFloat",
        "illuminance_ave"    : "WSFloat",
        "illuminance_max"    : "WSFloat",
        "illuminance_max_t"  : "WSDateTime",
        "uv_ave"             : "WSFloat",
        "uv_max"             : "WSInt",
        "uv_max_t"           : "WSDateTime"
    }
    key_list = tuple(conv.keys())


class MonthlyStore(CoreStore):
    """Stores monthly summary weather station data."""
    table = "monthly"
    conv = {
        "idx"                  : "WSDateTime",
        "start"                : "WSDateTime",
        "hum_out_ave"          : "WSFloat",
        "hum_out_min"          : "WSInt",
        "hum_out_min_t"        : "WSDateTime",
        "hum_out_max"          : "WSInt",
        "hum_out_max_t"        : "WSDateTime",
        "temp_out_ave"         : "WSFloat",
        "temp_out_min_lo"      : "WSFloat",
        "temp_out_min_lo_t"    : "WSDateTime",
        "temp_out_min_hi"      : "WSFloat",
        "temp_out_min_hi_t"    : "WSDateTime",
        "temp_out_min_ave"     : "WSFloat",
        "temp_out_max_lo"      : "WSFloat",
        "temp_out_max_lo_t"    : "WSDateTime",
        "temp_out_max_hi"      : "WSFloat",
        "temp_out_max_hi_t"    : "WSDateTime",
        "temp_out_max_ave"     : "WSFloat",
        "hum_in_ave"           : "WSFloat",
        "hum_in_min"           : "WSInt",
        "hum_in_min_t"         : "WSDateTime",
        "hum_in_max"           : "WSInt",
        "hum_in_max_t"         : "WSDateTime",
        "temp_in_ave"          : "WSFloat",
        "temp_in_min_lo"       : "WSFloat",
        "temp_in_min_lo_t"     : "WSDateTime",
        "temp_in_min_hi"       : "WSFloat",
        "temp_in_min_hi_t"     : "WSDateTime",
        "temp_in_min_ave"      : "WSFloat",
        "temp_in_max_lo"       : "WSFloat",
        "temp_in_max_lo_t"     : "WSDateTime",
        "temp_in_max_hi"       : "WSFloat",
        "temp_in_max_hi_t"     : "WSDateTime",
        "temp_in_max_ave"      : "WSFloat",
        "abs_pressure_ave"     : "WSFloat",
        "abs_pressure_min"     : "WSFloat",
        "abs_pressure_min_t"   : "WSDateTime",
        "abs_pressure_max"     : "WSFloat",
        "abs_pressure_max_t"   : "WSDateTime",
        "rel_pressure_ave"     : "WSFloat",
        "rel_pressure_min"     : "WSFloat",
        "rel_pressure_min_t"   : "WSDateTime",
        "rel_pressure_max"     : "WSFloat",
        "rel_pressure_max_t"   : "WSDateTime",
        "wind_ave"             : "WSFloat",
        "wind_gust"            : "WSFloat",
        "wind_gust_t"          : "WSDateTime",
        "wind_dir"             : "WSFloat",
        "rain"                 : "WSFloat",
        "rain_days"            : "WSInt",
        "illuminance_ave"      : "WSFloat",
        "illuminance_max_lo"   : "WSFloat",
        "illuminance_max_lo_t" : "WSDateTime",
        "illuminance_max_hi"   : "WSFloat",
        "illuminance_max_hi_t" : "WSDateTime",
        "illuminance_max_ave"  : "WSFloat",
        "uv_ave"               : "WSFloat",
        "uv_max_lo"            : "WSInt",
        "uv_max_lo_t"          : "WSDateTime",
        "uv_max_hi"            : "WSInt",
        "uv_max_hi_t"          : "WSDateTime",
        "uv_max_ave"           : "WSFloat"
    }
    key_list = tuple(conv.keys())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018-20  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload weather data to Open Weather Map.

`Open Weather Map`_ is a Latvian based IT company seeking to provide
affordable weather data.

* Create account: https://home.openweathermap.org/users/sign_up
* API: https://openweathermap.org/stations
* Additional dependency: http://docs.python-requests.org/
* Example ``weather.ini`` configuration::

    [openweathermap]
    api key = b1b15e88fa797225412429c1c50c122a1
    external id = SW1Aweather
    station name = Buck House
    lat = 51.501
    long = -0.142
    alt = 10
    station id = 583436dd9643a9000196b8d6

    [logged]
    services = ['openweathermap', 'underground']

Configuring pywws to use Open Weather Map is a bit more complicated than
with other services. Start by running the module to set some default
values in weather.ini (with no other pywws software running)::

    python -m pywws.service.openweathermap data_dir

After signing up and logging in to Open Weather Map visit the `API keys
page`_ and copy your default key to the ``api key`` entry in
weather.ini. The ``external id`` field is a single word name to identify
your station. You could use a name based on your post code, or maybe
your id from Weather Underground or CWOP. The ``station name`` is a
longer, human readable, name. I'm not sure what use Open Weather Map
makes of either of these. ``lat`` and ``long`` should be set to the
latitude and longitude of your station (in degrees) and ``alt`` to its
altitude in metres.

After setting (or changing) the above fields you need to "register" your
station with Open Weather Map. This is done by running the module with
the ``-r`` flag::

    python -m pywws.service.openweathermap -r -v data_dir

If you already have any stations registered with Open Weather Map this
will show you their details. You can select one of these existing
stations or register a new one. pywws then sends the configuration
values from weather.ini to Open Weather Map.

If this succeeds then Open Weather Map will have allocated a ``station
id`` value which pywws stores in weather.ini. All this complication is
to allow more than one station to be attached to one user's account.

.. _Open Weather Map: https://openweathermap.org/
.. _API keys page: https://home.openweathermap.org/api_keys

"""

from __future__ import absolute_import, unicode_literals

from contextlib import contextmanager
from datetime import timedelta
import json
import logging
import os
import sys

import requests

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.CatchupDataService):
    config = {
        'api key'     : ('', True,  None),
        'station id'  : ('', True,  'station_id'),
        'external id' : ('', False, None),
        'station name': ('', False, None),
        'lat'         : ('', False, None),
        'long'        : ('', False, None),
        'alt'         : ('', False, None),
        }
    logger = logger
    service_name = service_name
    template = """
#live#
#idx          "'dt'         : %s,"#
#temp_out     "'temperature': %.1f,"#
#wind_ave     "'wind_speed' : %.1f,"#
#wind_gust    "'wind_gust'  : %.1f,"#
#wind_dir     "'wind_deg'   : %.0f," "" "winddir_degrees(x)"#
#rel_pressure "'pressure'   : %.1f,"#
#hum_out      "'humidity'   : %.d,"#
#calc "rain_hour(data)" "'rain_1h': %.1f,"#
#calc "rain_24hr(data)" "'rain_24h': %.1f,"#
#calc "dew_point(data['temp_out'], data['hum_out'])" "'dew_point': %.1f,"#
"""

    @contextmanager
    def session(self):
        with requests.Session() as session:
            session.headers.update({'Content-Type': 'application/json'})
            session.params.update({'appid': self.params['api key']})
            yield session

    def upload_data(self, session, prepared_data={}):
        url = 'https://api.openweathermap.org/data/3.0/measurements'
        try:
            rsp = session.post(url, json=[prepared_data], timeout=60)
        except Exception as ex:
            return False, repr(ex)
        if rsp.status_code != 204:
            return False, 'http status: {:d} {:s}'.format(
                rsp.status_code, rsp.text)
        return True, 'OK'

    def register(self):
        import pprint

        self.check_params('external id', 'station name', 'lat', 'long', 'alt')
        url = 'https://api.openweathermap.org/data/3.0/stations'
        data = {
            'external_id': self.params['external id'],
            'name'       : self.params['station name'],
            'latitude'   : float(self.params['lat']),
            'longitude'  : float(self.params['long']),
            'altitude'   : float(self.params['alt']),
            }
        station_id = self.params['station id']
        idx = -1
        with self.session() as session:
            # get current stations
            try:
                rsp = session.get(url, timeout=60)
            except Exception as ex:
                print('exception', repr(ex))
                return
            stations = rsp.json()
            if stations:
                print('The following stations are registered to your account')
                for i, station in enumerate(stations):
                    if station['id'] == station_id:
                        idx = i
                        print('Number:', i, '\t\t\t\t<- current station')
                    else:
                        print('Number:', i)
                    pprint.pprint(station)
                if sys.version_info[0] >= 3:
                    input_ = input
                else:
                    input_ = raw_input
                i = input_('Please enter number of station to use, or N' +
                           ' to register a new station: ')
                if i in ('N', 'n'):
                    idx = -1
                    station_id = None
                else:
                    idx = int(i)
                    station_id = stations[idx]['id']
                for i, station in enumerate(stations):
                    if i == idx:
                        continue
                    yn = input_('Would you like to delete station number' +
                                ' {} and all its data (Y/N)? '.format(i))
                    if yn in ('Y', 'y'):
                        try:
                            session.delete(
                                url + '/' + station['id'], timeout=60)
                        except Exception as ex:
                            print('exception', repr(ex))
                            return
            if station_id:
                # update existing station
                logger.debug('Udating station id ' + station_id)
                url += '/' + station_id
                try:
                    rsp = session.put(url, json=data, timeout=60)
                except Exception as ex:
                    print('exception', repr(ex))
                    return
                rsp = rsp.json()
                logger.debug('response: ' + repr(rsp))
            else:
                # create new station
                logger.debug('Creating new station')
                try:
                    rsp = session.post(url, json=data, timeout=60)
                except Exception as ex:
                    print('exception', repr(ex))
                    return
                rsp = rsp.json()
                logger.debug('response: ' + repr(rsp))
            for key in 'id', 'ID':
                if key in rsp:
                    self.context.params.set(
                        service_name, 'station id', rsp[key])


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Regenerate hourly and daily summary data.

This module can also be run with the ``pywws-reprocess`` command. ::
%s
This module recreates the calibrated, hourly, daily and monthly
summary data that is created by the :py:mod:`pywws.process` module. It
should be run whenever you upgrade to a newer version of pywws (if the
summary data format has changed), change your calibration module or
alter your pressure offset.

The ``-u`` (or ``--update``) option is a special case. It should be
used when upgrading from any pywws version earlier than 14.02.dev1143.
Unlike normal reprocessing, the ``-u`` option changes your raw data.
You are advised to backup your data before using the ``-u`` option.

"""

from __future__ import absolute_import, print_function

__docformat__ = "restructuredtext en"
__usage__ = """
 usage: %s [options] data_dir
 options are:
  -h | --help     display this help
  -u | --update   update status on old data to include bits from wind_dir byte
  -v | --verbose  increase number of informative messages
 data_dir is the root directory of the weather data
"""
__doc__ %= __usage__ % ('python -m pywws.reprocess')

import getopt
import logging
import os
import sys

import pywws.logger
import pywws.process
import pywws.storage

logger = logging.getLogger(__name__)


def reprocess(data_dir, update):
    with pywws.storage.pywws_context(data_dir) as context:
        if update:
            logger.warning("Updating status to detect invalid wind_dir")
            count = 0
            raw_data = context.raw_data
            for data in raw_data[:]:
                count += 1
                idx = data['idx']
                if count % 10000 == 0:
                    logger.info("update: %s", idx.isoformat(' '))
                elif count % 500 == 0:
                    logger.debug("update: %s", idx.isoformat(' '))
                if data['wind_dir'] is not None and (data['wind_dir'] & 0x80):
                    data['wind_dir'] = None
                    raw_data[idx] = data
            raw_data.flush()
        # delete old format summary files
        logger.warning('Deleting old summaries')
        context.calib_data.clear()
        context.hourly_data.clear()
        context.daily_data.clear()
        context.monthly_data.clear()
        # create data summaries
        logger.warning('Generating hourly and daily summaries')
        pywws.process.process_data(context)
    return 0


def main(argv=None):
    if argv is None:
        argv = sys.argv
    usage = (__usage__ % (argv[0])).strip()
    try:
        opts, args = getopt.getopt(
            argv[1:], "huv", ['help', 'update', 'verbose'])
    except getopt.error as msg:
        print('Error: %s\n' % msg, file=sys.stderr)
        print(usage, file=sys.stderr)
        return 1
    # process options
    update = False
    verbose = 0
    for o, a in opts:
        if o in ('-h', '--help'):
            print(__doc__.split('\n\n')[0])
            print(usage)
            return 0
        elif o in ('-u', '--update'):
            update = True
        elif o in ('-v', '--verbose'):
            verbose += 1
    # check arguments
    if len(args) != 1:
        print('Error: 1 argument required\n', file=sys.stderr)
        print(usage, file=sys.stderr)
        return 2
    pywws.logger.setup_handler(verbose)
    data_dir = args[0]
    return reprocess(data_dir, update)


if __name__ == "__main__":
    sys.exit(main())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload weather data to UK Met Office "WOW".

The UK Met Office runs a `Weather Observations Website`_ (WOW) that
displays readings from amateur and official weather stations. This
module uploads data to it from pywws. You can upload "logged" or "live"
data (or both). The module ensures there is at least 5 minutes between
each reading as required by the API.

* Create account: https://register.metoffice.gov.uk/WaveRegistrationClient/public/newaccount.do?service=weatherobservations
* API: http://wow.metoffice.gov.uk/support/dataformats#automatic
* Additional dependency: http://docs.python-requests.org/
* Example ``weather.ini`` configuration::

    [metoffice]
    site id = 12345678
    aws pin = 987654

    [logged]
    services = ['metoffice', 'underground']

    [live]
    services = ['metoffice', 'underground']

Note that a ``site id`` allocated since June 2016 will probably look
like ``6a571450-df53-e611-9401-0003ff5987fd``.

.. _Weather Observations Website: http://wow.metoffice.gov.uk/home

"""

from __future__ import absolute_import, unicode_literals

from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import sys

import requests

import pywws
from pywws.conversions import rain_inch
from pywws.process import get_day_end_hour
import pywws.service
from pywws.timezone import timezone

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.CatchupDataService):
    config = {
        'site id': ('', True, 'siteid'),
        'aws pin': ('', True, 'siteAuthenticationKey'),
        }
    fixed_data = {'softwaretype': 'pywws-' + pywws.__version__}
    interval = timedelta(seconds=300)
    logger = logger
    service_name = service_name
    template = """
#live#
#idx          "'dateutc'     : '%Y-%m-%d %H:%M:%S',"#
#wind_dir     "'winddir'     : '%.0f'," "" "winddir_degrees(x)"#
#wind_ave     "'windspeedmph': '%.2f'," "" "wind_mph(x)"#
#wind_gust    "'windgustmph' : '%.2f'," "" "wind_mph(x)"#
#hum_out      "'humidity'    : '%.d',"#
#temp_out     "'tempf'       : '%.1f'," "" "temp_f(x)"#
#rel_pressure "'baromin'     : '%.4f'," "" "pressure_inhg(x)"#
#calc "rain_inch(self.rain_rate(data))"
              "'rainin'      : '%.4f',"#
#calc "rain_inch(self.rain_day_local(data))"
              "'dailyrainin' : '%.4f',"#
#calc "temp_f(dew_point(data['temp_out'], data['hum_out']))"
              "'dewptf'      : '%.1f',"#
"""

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # initialise rain history
        last_update = context.calib_data.nearest(self.last_update)
        self.last_rain = context.calib_data[last_update]['rain']
        # add rain functions to templater
        self.templater.rain_rate = self.rain_rate
        self.templater.rain_day_local = self.rain_day_local

    @contextmanager
    def session(self):
        with requests.Session() as session:
            yield session

    def rain_rate(self, data):
        # compute rain since last upload
        result = max(data['rain'] - self.last_rain, 0.0)
        self.last_rain = data['rain']
        return result

    def rain_day_local(self, data):
        # compute rain since day start
        day_end_hour, use_dst = get_day_end_hour(self.context.params)
        day_start = timezone.local_replace(
            data['idx'], use_dst=use_dst, hour=day_end_hour, minute=0, second=0)
        day_start = self.context.calib_data.nearest(day_start)
        return max(
            data['rain'] - self.context.calib_data[day_start]['rain'], 0.0)

    def valid_data(self, data):
        return any([data[x] is not None for x in (
            'wind_dir', 'wind_ave', 'wind_gust', 'hum_out', 'temp_out',
            'rel_pressure')])

    def upload_data(self, session, prepared_data={}):
        try:
            rsp = session.get('http://wow.metoffice.gov.uk/automaticreading',
                              params=prepared_data, timeout=60)
        except Exception as ex:
            return False, repr(ex)
        if rsp.status_code == 429:
            # UK Met Office server uses 429 to signal duplicate data
            return True, 'repeated data {:s}'.format(prepared_data['dateutc'])
        if rsp.status_code != 200:
            return False, 'http status: {:d}'.format(rsp.status_code)
        rsp = rsp.json()
        if rsp:
            return True, 'server response "{!r}"'.format(rsp)
        return True, 'OK'


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-19  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Get data from WH1080/WH3080 compatible weather stations.

Derived from wwsr.c by Michael Pendec (michael.pendec@gmail.com),
wwsrdump.c by Svend Skafte (svend@skafte.net), modified by Dave Wells,
and other sources.

Introduction
------------

This is the module that actually talks to the weather station base
unit. I don't have much understanding of USB, so copied a lot from
Michael Pendec's C program wwsr.

The weather station memory has two parts: a "fixed block" of 256 bytes
and a circular buffer of 65280 bytes. As each weather reading takes 16
bytes the station can store 4080 readings, or 14 days of 5-minute
interval readings. (The 3080 type stations store 20 bytes per reading,
so store a maximum of 3264.) As data is read in 32-byte chunks, but
each weather reading is 16 or 20 bytes, a small cache is used to
reduce USB traffic. The caching behaviour can be over-ridden with the
``unbuffered`` parameter to ``get_data`` and ``get_raw_data``.

Decoding the data is controlled by the static dictionaries
``_reading_format``, ``lo_fix_format`` and ``fixed_format``. The keys
are names of data items and the values can be an ``(offset, class,
kwds)`` tuple or another dictionary. So, for example, the
_reading_format dictionary entry ``'rain' : (13, WSFloat, {'signed':
False, 'scale': 0.3})`` means that the rain value is an unsigned short
(two bytes), 13 bytes from the start of the block, and should be
multiplied by 0.3 to get a useful value.

The use of nested dictionaries in the ``fixed_format`` dictionary
allows useful subsets of data to be decoded. For example, to decode
the entire block ``get_fixed_block`` is called with no parameters::

  ws = pywws.weatherstation.WeatherStation()
  print(ws.get_fixed_block())

To get the stored minimum external temperature, ``get_fixed_block`` is
called with a sequence of keys::

  ws = pywws.weatherstation.WeatherStation()
  print(ws.get_fixed_block(['min', 'temp_out', 'val']))

Often there is no requirement to read and decode the entire fixed
block, as its first 64 bytes contain the most useful data: the
interval between stored readings, the buffer address where the current
reading is stored, and the current date & time. The
``get_lo_fix_block`` method provides easy access to these.

For more examples of using the pywws.weatherstation module, see the
pywws.testweatherstation module.

Detailed API
------------

"""

from __future__ import absolute_import

__docformat__ = "restructuredtext en"

from ast import literal_eval
from datetime import datetime
import logging
import sys
import time

for module_name in ('device_libusb1', 'device_pyusb1', 'device_pyusb',
                    'device_ctypes_hidapi', 'device_cython_hidapi'):
    try:
        module = __import__(module_name, globals(), locals(), level=1)
        USBDevice = getattr(module, 'USBDevice')
        break
    except ImportError:
        pass
else:
    raise ImportError('No USB library found')

logger = logging.getLogger(__name__)


class WSBits(dict):
    @staticmethod
    def from_int(value, keys):
        # convert byte to list of 8 booleans
        mask = 1
        values = []
        for i in range(8):
            values.append(value & mask != 0)
            mask = mask << 1
        # merge with keys to make a dict
        return WSBits(zip(keys, values))

    @staticmethod
    def from_raw(raw, pos, keys=[]):
        return WSBits.from_int(raw[pos], keys)

    # don't display unknown bits
    def __repr__(self):
        result = {}
        for key in self:
            if self[key] or not key.startswith('bit'):
                result[key] = self[key]
        return repr(result)


class WSStatus(WSBits):
    keys = ('bit0', 'bit1', 'bit2', 'bit3', 'bit4', 'bit5',
            'lost_connection', 'rain_overflow')

    @classmethod
    def from_raw(cls, raw, pos):
        return WSStatus(WSBits.from_int(raw[pos], cls.keys))

    # convert to stringified int
    def to_csv(self):
        mask = 1
        value = 0
        for key in self.keys:
            if self[key]:
                value += mask
            mask = mask << 1
        return str(value)

    # convert from stringified int
    @classmethod
    def from_csv(cls, value):
        if not value:
            return None
        return WSStatus(WSBits.from_int(int(value), cls.keys))


class WSInt(int):
    @staticmethod
    def from_1(raw, pos, signed=False):
        # decode one byte to an int
        value = raw[pos]
        if value == 0xFF:
            return None
        if signed and value >= 0x80:
            value = 0x80 - value
        return WSInt(value)

    @staticmethod
    def wind_dir(raw, pos):
        # decode one byte to an int
        value = raw[pos]
        # if bit 7 is 1, value is invalid
        if value & 0x80:
            return None
        return WSInt(value)

    @staticmethod
    def from_2(raw, pos, signed=False):
        # decode two bytes to an int
        value = raw[pos] + (raw[pos+1] << 8)
        if value == 0xFFFF:
            return None
        if signed and value >= 0x8000:
            value = 0x8000 - value
        return WSInt(value)

    @staticmethod
    def from_3(raw, pos, signed=False):
        # decode three bytes to an int
        value = raw[pos] + (raw[pos+1] << 8) + (raw[pos+2] << 16)
        if value == 0xFFFFFF:
            return None
        if signed and value >= 0x800000:
            value = 0x800000 - value
        return WSInt(value)


def _nibble_value(raw, base_shift, nibble_pos=None, nibble_high=False):
    if nibble_pos is None:
        return 0
    mask = 0x0F
    shift = base_shift
    if nibble_high:
        mask = 0xF0
        shift = base_shift-4
    return ((raw[nibble_pos] & mask) << shift)


class WSFloat(float):
    @staticmethod
    def from_1(raw, pos, signed=False, scale=1.0, nibble_pos=None, nibble_high=False):
        # decode one byte to an int
        value = WSInt.from_1(raw, pos, signed=signed)
        if value is None:
            return None
        value += _nibble_value(raw, 8, nibble_pos=nibble_pos, nibble_high=nibble_high)
        # convert to float
        return WSFloat(float(value) * scale)

    @staticmethod
    def from_2(raw, pos, signed=False, scale=1.0, nibble_pos=None, nibble_high=False):
        # decode two bytes to an int
        value = WSInt.from_2(raw, pos, signed=signed)
        if value is None:
            return None
        value += _nibble_value(raw, 16, nibble_pos=nibble_pos, nibble_high=nibble_high)
        # convert to float
        return WSFloat(float(value) * scale)

    @staticmethod
    def from_3(raw, pos, signed=False, scale=1.0):
        # decode three bytes to an int
        value = WSInt.from_3(raw, pos, signed=signed)
        if value is None:
            return None
        # convert to float
        return WSFloat(float(value) * scale)

    # don't display excessive precision
    def __str__(self):
        return '{:.12g}'.format(self)

    def __repr__(self):
        return '{:.12g}'.format(self)


def _bcd_decode(byte):
    hi = (byte & 0xF0) >> 4
    lo = byte & 0x0F
    return (hi * 10) + lo


class WSTime(str):
    @staticmethod
    def from_raw(raw, pos):
        hour = _bcd_decode(raw[pos])
        minute = _bcd_decode(raw[pos+1])
        try:
            return WSTime('{:02d}:{:02d}'.format(hour, minute))
        except ValueError:
            return None


class WSDateTime(datetime):
    # only save string representation in 'fixed block' in status.ini
    def __repr__(self):
        return repr(self.strftime('%Y-%m-%d %H:%M:%S'))

    def to_csv(self):
        return self.strftime('%Y-%m-%d %H:%M:%S')

    @staticmethod
    def from_csv(date_string):
        return WSDateTime(*map(int, (date_string[0:4],
                                     date_string[5:7],
                                     date_string[8:10],
                                     date_string[11:13],
                                     date_string[14:16],
                                     date_string[17:19])))

    @staticmethod
    def from_raw(raw, pos):
        year = _bcd_decode(raw[pos])
        month = _bcd_decode(raw[pos+1])
        day = _bcd_decode(raw[pos+2])
        hour = _bcd_decode(raw[pos+3])
        minute = _bcd_decode(raw[pos+4])
        try:
            return WSDateTime(year + 2000, month, day, hour, minute)
        except ValueError:
            return None


def _decode(raw, format_):
    if not raw:
        return None
    if isinstance(format_, dict):
        result = {}
        for key, value in format_.items():
            result[key] = _decode(raw, value)
    else:
        pos, factory, kwds = format_
        result = factory(raw, pos, **kwds)
    return result


class CUSBDrive(object):
    """Low level interface to weather station via USB.

    Loosely modeled on a C++ class obtained from
    http://site.ambientweatherstore.com/easyweather/ws_1080_2080_protocol.zip.
    I don't know the provenance of this, but it looks as if it may
    have come from the manufacturer.

    """
    EndMark          = 0x20
    ReadCommand      = 0xA1
    WriteCommand     = 0xA0
    WriteCommandWord = 0xA2

    def __init__(self):
        logger.info('using %s', USBDevice.__module__)
        self.dev = USBDevice(0x1941, 0x8021)

    def read_block(self, address):
        """Read 32 bytes from the weather station.

        If the read fails for any reason, :obj:`None` is returned.

        :param address: address to read from.

        :type address: int

        :return: the data from the weather station.

        :rtype: list(int)

        """
        buf = [
            self.ReadCommand,
            address // 256,
            address % 256,
            self.EndMark,
            self.ReadCommand,
            address // 256,
            address % 256,
            self.EndMark,
            ]
        if not self.dev.write_data(buf):
            return None
        return self.dev.read_data(32)

    def write_byte(self, address, data):
        """Write a single byte to the weather station.

        :param address: address to write to.

        :type address: int

        :param data: the value to write.

        :type data: int

        :return: success status.

        :rtype: bool

        """
        buf = [
            self.WriteCommandWord,
            address // 256,
            address % 256,
            self.EndMark,
            self.WriteCommandWord,
            data,
            0,
            self.EndMark,
            ]
        if not self.dev.write_data(buf):
            return False
        buf = self.dev.read_data(8)
        if buf is None:
            return False
        for byte in buf:
            if byte != 0xA5:
                return False
        return True


class DriftingClock(object):
    def __init__(self, name, status, period, margin):
        self.name = name
        self.status = status
        self.period = period
        self.margin = margin
        if self.status:
            self.clock = literal_eval(
                self.status.get('clock', self.name, 'None'))
            self.drift = literal_eval(
                self.status.get('clock', self.name + ' drift', '0.0'))
        else:
            self.clock = None
            self.drift = 0.0
        self._set_real_period()
        self.old_clock = self.clock

    def _set_real_period(self):
        self._real_period = self.period * (1.0 + (self.drift / (24.0 * 3600.0)))

    def before(self, now):
        if not self.clock:
            return None
        error = (now - self.clock) % self._real_period
        return now - error

    def avoid(self):
        if not self.clock:
            return 1000.0
        now = time.time()
        phase = now - self.clock
        if phase > 24 * 3600:
            # clock was last measured a day ago, so reset it
            self.clock = None
            return 1000.0
        return (self.margin - phase) % self._real_period

    def set_clock(self, now):
        if self.clock:
            diff = (now - self.clock) % self._real_period
            if diff < 2.0 or diff > self._real_period - 2.0:
                return
            logger.error('unexpected %s clock change', self.name)
        self.clock = now
        logger.warning('setting %s clock %g', self.name, now % self.period)
        if self.status:
            self.status.set('clock', self.name, str(self.clock))
        if self.old_clock:
            diff = now - self.old_clock
            if diff < 8 * 3600:
                # drift measurement needs more than 8 hours gap
                return
            drift = diff % self.period
            if drift > self.period / 2:
                drift -= self.period
            drift = (float(drift) * 24.0 * 3600.0 / float(diff))
            self.drift += max(min(drift - self.drift, 3.0), -3.0) / 4.0
            self._set_real_period()
            logger.warning(
                '%s clock drift %g %g', self.name, drift, self.drift)
            if self.status:
                self.status.set(
                    'clock', '%s drift' % self.name, str(self.drift))
        self.old_clock = self.clock

    def invalidate(self):
        self.clock = None


class WeatherStation(object):
    """Class that represents the weather station to user program."""
    # minimum interval between polling for data change
    min_pause = 0.5
    # margin of error for various decisions
    margin = (min_pause * 2.0) - 0.1

    def __init__(self, context=None):
        """Connect to weather station and prepare to read data."""
        # create basic IO object
        self.cusb = CUSBDrive()
        # init variables
        if context:
            self.status = context.status
            self.avoid = float(
                context.params.get('config', 'usb activity margin', '3.0'))
            self.avoid = max(self.avoid, 0.0)
            self.ws_type = context.params.get('config', 'ws type', 'Unknown')
        else:
            self.status = None
            self.avoid = 3.0
            self.ws_type = '1080'
        self._fixed_block = None
        self._data_block = None
        self._data_pos = None
        self._current_ptr = None
        self._station_clock = DriftingClock(
            'station', self.status, 60, self.avoid)
        self._sensor_clock = DriftingClock(
            'sensor', self.status, 48, self.avoid)
        if self.ws_type == '3080':
            self._solar_clock = DriftingClock(
                'solar', self.status, 60, self.avoid)
        else:
            self._solar_clock = None
        self.last_status = {}
        # check ws_type
        if self.ws_type not in ('1080', '3080'):
            if self.get_fixed_block(['lux_wm2_coeff']) == 0.0:
                guess = '1080'
            else:
                guess = '3080'
            raise ValueError("""
Unknown weather station type. Please edit weather.ini and set 'ws type' to
'1080' or '3080', as appropriate.
Your station is probably a '{:s}' type.
""".format(guess))

    def live_data(self, logged_only=False):
        # There are two things we want to synchronise to - the data is
        # updated every 48 seconds and the address is incremented
        # every 5 minutes (or 10, 15, ..., 30). Rather than getting
        # data every second or two, we sleep until one of the above is
        # due. (During initialisation we get data every half second
        # anyway.)
        read_period = self.get_fixed_block(['read_period'])
        logger.debug('read period %d', read_period)
        log_interval = float(read_period * 60)
        live_interval = 48.0
        old_ptr = self.current_pos()
        old_data = self.get_data(old_ptr, unbuffered=True)
        now = time.time()
        next_live = self._sensor_clock.before(now + live_interval)
        if next_live:
            now = next_live - live_interval
        else:
            now -= live_interval
        last_log = now - (old_data['delay'] * 60.0)
        next_log = self._station_clock.before(last_log + log_interval)
        ptr_time = 0
        data_time = 0
        not_logging = False
        while True:
            # sleep until just before next reading is due
            now = time.time()
            advance = now + max(self.avoid, self.min_pause) + self.min_pause
            if next_live:
                pause = next_live - advance
            else:
                pause = self.min_pause
            if not_logging:
                pass
            elif next_log:
                pause = min(pause, next_log - advance)
            elif old_data['delay'] >= read_period - 1:
                pause = self.min_pause
            if (self._solar_clock and self._solar_clock.clock is None
                    and old_data['illuminance'] is not None):
                pause = self.min_pause
            pause = max(pause, self.min_pause)
            logger.debug(
                'delay %s, pause %g', str(old_data['delay']), pause)
            time.sleep(pause)
            # get new pointer
            last_ptr_time = ptr_time
            new_ptr = self.current_pos()
            ptr_time = time.time()
            # get new data
            last_data_time = data_time
            new_data = self.get_data(old_ptr, unbuffered=True)
            data_time = time.time()
            # when ptr changes, internal sensor data gets updated
            if new_ptr != old_ptr:
                for key in ('hum_in', 'temp_in', 'abs_pressure'):
                    old_data[key] = new_data[key]
            # log any change of status
            if new_data['status'] != self.last_status:
                logger.warning('status %s', str(new_data['status']))
            self.last_status = new_data['status']
            if (new_data['status']['lost_connection'] and not
                    old_data['status']['lost_connection']):
                # 'lost connection' decision can happen at any time
                old_data = new_data
            # has data changed?
            if any(new_data[key] != old_data[key] for key in (
                    'hum_in', 'temp_in', 'hum_out', 'temp_out',
                    'abs_pressure', 'wind_ave', 'wind_gust', 'wind_dir',
                    'rain', 'status')):
                logger.debug('live_data new data')
                if data_time - last_data_time < self.margin:
                    # data has just changed, so definitely at a 48s update time
                    self._sensor_clock.set_clock(data_time)
                elif next_live and data_time < next_live - self.margin:
                    logger.warning(
                        'live_data lost sync %g', data_time - next_live)
                    logger.warning('old data %s', str(old_data))
                    logger.warning('new data %s', str(new_data))
                    self._sensor_clock.invalidate()
                next_live = self._sensor_clock.before(data_time + self.margin)
                if next_live:
                    if not logged_only:
                        result = dict(new_data)
                        result['idx'] = datetime.utcfromtimestamp(int(next_live))
                        yield result, old_ptr, False
                    next_live += live_interval
                    if not_logging:
                        # simulate logging if station is not logging
                        if not next_log:
                            next_log = next_live
                        if next_live > next_log:
                            result = dict(new_data)
                            result['idx'] = datetime.utcfromtimestamp(int(next_log))
                            yield result, old_ptr, True
                            next_log += log_interval
            elif next_live and data_time > next_live + 6.0:
                logger.info('live_data missed')
                next_live += live_interval
            # has solar data changed?
            elif self._solar_clock and (
                    new_data['illuminance'] != old_data['illuminance'] or
                    new_data['uv'] != old_data['uv']):
                logger.debug('live_data new solar data')
                if data_time - last_data_time < self.margin:
                    # data has just changed, so at a solar update time
                    self._solar_clock.set_clock(data_time)
            old_data = new_data
            # has ptr changed?
            if new_ptr != old_ptr:
                logger.info('live_data new ptr: %06x', new_ptr)
                if not_logging:
                    logger.error('station is logging data')
                    not_logging = False
                last_log = ptr_time - self.margin
                if ptr_time - last_ptr_time < self.margin:
                    # pointer has just changed, so definitely at a logging time
                    self._station_clock.set_clock(ptr_time)
                elif next_log:
                    if ptr_time < next_log - self.margin:
                        logger.warning(
                            'live_data lost log sync %g', ptr_time - next_log)
                        self._station_clock.invalidate()
                else:
                    logger.info('missed ptr change time')
                if read_period > new_data['delay']:
                    read_period = new_data['delay']
                    logger.warning('reset read period %d', read_period)
                    log_interval = float(read_period * 60)
                result = dict(new_data)
                next_log = self._station_clock.before(ptr_time + self.margin)
                if next_log:
                    result['idx'] = datetime.utcfromtimestamp(int(next_log))
                    next_log += log_interval
                else:
                    # use best guess of logging time
                    result['idx'] = datetime.utcfromtimestamp(
                        int(ptr_time - (self.avoid / 2)))
                yield result, old_ptr, True
                old_ptr = new_ptr
                old_data['delay'] = 0
                data_time = 0
            elif ptr_time > last_log + log_interval + 180.0:
                # if station stops logging data, don't keep reading
                # USB until it locks up
                logger.error('station is not logging data')
                not_logging = True
            elif next_log and ptr_time > next_log + 6.0:
                logger.warning('live_data log extended')
                next_log += 60.0

    def inc_ptr(self, ptr):
        """Get next circular buffer data pointer."""
        result = ptr + self.reading_len[self.ws_type]
        if result >= 0x10000:
            result = self.data_start
        return result

    def dec_ptr(self, ptr):
        """Get previous circular buffer data pointer."""
        result = ptr - self.reading_len[self.ws_type]
        if result < self.data_start:
            result = 0x10000 - self.reading_len[self.ws_type]
        return result

    def get_raw_data(self, ptr, unbuffered=False):
        """Get raw data from circular buffer.

        If unbuffered is false then a cached value that was obtained
        earlier may be returned."""
        if unbuffered:
            self._data_pos = None
        # round down ptr to a 'block boundary'
        idx = ptr - (ptr % 0x20)
        ptr -= idx
        count = self.reading_len[self.ws_type]
        if self._data_pos == idx:
            # cache contains useful data
            result = self._data_block[ptr:ptr + count]
            if len(result) >= count:
                return result
        else:
            result = list()
        if ptr + count > 0x20:
            # need part of next block, which may be in cache
            if self._data_pos != idx + 0x20:
                self._data_pos = idx + 0x20
                self._data_block = self._read_block(self._data_pos)
            result += self._data_block[0:ptr + count - 0x20]
            if len(result) >= count:
                return result
        # read current block
        self._data_pos = idx
        self._data_block = self._read_block(self._data_pos)
        result = self._data_block[ptr:ptr + count] + result
        return result

    def get_data(self, ptr, unbuffered=False):
        """Get decoded data from circular buffer.

        If unbuffered is false then a cached value that was obtained
        earlier may be returned."""
        result = _decode(self.get_raw_data(ptr, unbuffered),
                         self._reading_format[self.ws_type])
        return result

    def current_pos(self):
        """Get circular buffer location where current data is being written."""
        new_ptr = _decode(
            self._read_fixed_block(0x0020), self.lo_fix_format['current_pos'])
        if new_ptr == self._current_ptr:
            return self._current_ptr
        if self._current_ptr and new_ptr != self.inc_ptr(self._current_ptr):
            logger.error(
                'unexpected ptr change %06x -> %06x', self._current_ptr, new_ptr)
        self._current_ptr = new_ptr
        return self._current_ptr

    def get_raw_fixed_block(self, unbuffered=False):
        """Get the raw "fixed block" of settings and min/max data."""
        if unbuffered or not self._fixed_block:
            self._fixed_block = self._read_fixed_block()
        return self._fixed_block

    def get_fixed_block(self, keys=[], unbuffered=False):
        """Get the decoded "fixed block" of settings and min/max data.

        A subset of the entire block can be selected by keys."""
        if unbuffered or not self._fixed_block:
            self._fixed_block = self._read_fixed_block()
        format = self.fixed_format
        # navigate down list of keys to get to wanted data
        for key in keys:
            format = format[key]
        return _decode(self._fixed_block, format)

    def _wait_for_station(self):
        # avoid times when station is writing to memory
        while True:
            pause = min(self._station_clock.avoid(), self._sensor_clock.avoid())
            if self._solar_clock:
                pause = min(pause, self._solar_clock.avoid())
            if pause >= self.avoid * 2.0:
                return
            logger.debug('avoid %s', str(pause))
            time.sleep(pause)

    def _read_block(self, ptr, retry=True):
        # Read block repeatedly until it's stable. This avoids getting corrupt
        # data when the block is read as the station is updating it.
        old_block = None
        while True:
            self._wait_for_station()
            new_block = self.cusb.read_block(ptr)
            if new_block:
                if (new_block == old_block) or not retry:
                    break
                if old_block:
                    logger.debug('_read_block changing %06x', ptr)
                old_block = new_block
        return new_block

    def _read_fixed_block(self, hi=0x0100):
        result = []
        for mempos in range(0x0000, hi, 0x0020):
            result += self._read_block(mempos)
        return result

    def _write_byte(self, ptr, value):
        self._wait_for_station()
        if not self.cusb.write_byte(ptr, value):
            raise IOError('_write_byte failed')

    def write_data(self, data):
        """Write a set of single bytes to the weather station. Data must be an
        array of (ptr, value) pairs."""
        # send data
        for ptr, value in data:
            self._write_byte(ptr, value)
        # set 'data changed'
        self._write_byte(self.fixed_format['data_changed'][0], 0xAA)
        # wait for station to clear 'data changed'
        while True:
            ack = _decode(
                self._read_fixed_block(0x0020), self.fixed_format['data_changed'])
            if ack == 0:
                break
            logger.debug('write_data waiting for ack')
            time.sleep(6)

    # Tables of "meanings" for raw weather station data. Each key
    # specifies an (offset, factory, kwds) tuple that is understood
    # by _decode.
    # depends on weather station type
    _reading_format = {}
    _reading_format['1080'] = {
        'delay'        : (0, WSInt.from_1, {'signed': False}),
        'hum_in'       : (1, WSInt.from_1, {'signed': False}),
        'temp_in'      : (2, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
        'hum_out'      : (4, WSInt.from_1, {'signed': False}),
        'temp_out'     : (5, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
        'abs_pressure' : (7, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
        'wind_ave'     : (9, WSFloat.from_1, {'scale': 0.1, 'nibble_pos':11, 'nibble_high':False}),
        'wind_gust'    : (10, WSFloat.from_1, {'scale': 0.1, 'nibble_pos':11, 'nibble_high':True}),
        'wind_dir'     : (12, WSInt.wind_dir, {}),
        'rain'         : (13, WSFloat.from_2, {'signed': False, 'scale': 0.3}),
        'status'       : (15, WSStatus.from_raw, {}),
        }
    _reading_format['3080'] = {
        'illuminance' : (16, WSFloat.from_3, {'signed': False, 'scale': 0.1}),
        'uv'          : (19, WSInt.from_1, {'signed': False}),
        }
    _reading_format['3080'].update(_reading_format['1080'])
    lo_fix_format = {
        'read_period'   : (16, WSInt.from_1, {'signed': False}),
        'settings_1'    : (17, WSBits.from_raw, {'keys': (
            'temp_in_F', 'temp_out_F', 'rain_in', 'bit3', 'bit4',
            'pressure_hPa', 'pressure_inHg', 'pressure_mmHg')}),
        'settings_2'    : (18, WSBits.from_raw, {'keys': (
            'wind_mps', 'wind_kmph', 'wind_knot', 'wind_mph', 'wind_bft',
            'bit5', 'bit6', 'bit7')}),
        'display_1'     : (19, WSBits.from_raw, {'keys': (
            'pressure_rel', 'wind_gust', 'clock_12hr', 'date_mdy',
            'time_scale_24', 'show_year', 'show_day_name', 'alarm_time')}),
        'display_2'     : (20, WSBits.from_raw, {'keys': (
            'temp_out_temp', 'temp_out_chill', 'temp_out_dew', 'rain_hour',
            'rain_day', 'rain_week', 'rain_month', 'rain_total')}),
        'alarm_1'       : (21, WSBits.from_raw, {'keys': (
            'bit0', 'time', 'wind_dir', 'bit3', 'hum_in_lo', 'hum_in_hi',
            'hum_out_lo', 'hum_out_hi')}),
        'alarm_2'       : (22, WSBits.from_raw, {'keys': (
            'wind_ave', 'wind_gust', 'rain_hour', 'rain_day',
            'pressure_abs_lo', 'pressure_abs_hi',
            'pressure_rel_lo', 'pressure_rel_hi')}),
        'alarm_3'       : (23, WSBits.from_raw, {'keys': (
            'temp_in_lo', 'temp_in_hi', 'temp_out_lo', 'temp_out_hi',
            'wind_chill_lo', 'wind_chill_hi', 'dew_point_lo', 'dew_point_hi')}),
        'timezone'      : (24, WSInt.from_1, {'signed': True}),
        'unknown_01'    : (25, WSInt.from_1, {'signed': False}),
        'data_changed'  : (26, WSInt.from_1, {'signed': False}),
        'data_count'    : (27, WSInt.from_2, {'signed': False}),
        'display_3'     : (29, WSBits.from_raw, {'keys': (
            'illuminance_fc', 'alarm_illuminance_hi', 'alarm_uv_hi', 'bit3', 'bit4', 'illuminance_wm2', 'bit6',
            'bit7')}),
        'current_pos'   : (30, WSInt.from_2, {'signed': False}),
        }
    fixed_format = {
        'magic_0'       : (0,  WSInt.from_1, {'signed': False}),
        'magic_1'       : (1,  WSInt.from_1, {'signed': False}),
        'rain_factor_raw': (2,  WSFloat.from_2, {'signed': False}),
        'wind_factor_raw': (4,  WSFloat.from_2, {'signed': False}),
        'rel_pressure'  : (32, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
        'abs_pressure'  : (34, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
        'lux_wm2_coeff' : (36, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
        'date_time'     : (43, WSDateTime.from_raw, {}),
        'unknown_18'    : (97, WSInt.from_1, {'signed': False}),
        'alarm'         : {
            'hum_in'        : {'hi' : (48, WSInt.from_1, {'signed': False}),
                               'lo' : (49, WSInt.from_1, {'signed': False})},
            'temp_in'       : {'hi' : (50, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'lo' : (52, WSFloat.from_2, {'signed': True, 'scale': 0.1})},
            'hum_out'       : {'hi' : (54, WSInt.from_1, {'signed': False}),
                               'lo' : (55, WSInt.from_1, {'signed': False})},
            'temp_out'      : {'hi' : (56, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'lo' : (58, WSFloat.from_2, {'signed': True, 'scale': 0.1})},
            'windchill'     : {'hi' : (60, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'lo' : (62, WSFloat.from_2, {'signed': True, 'scale': 0.1})},
            'dewpoint'      : {'hi' : (64, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'lo' : (66, WSFloat.from_2, {'signed': True, 'scale': 0.1})},
            'abs_pressure'  : {'hi' : (68, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'lo' : (70, WSFloat.from_2, {'signed': False, 'scale': 0.1})},
            'rel_pressure'  : {'hi' : (72, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'lo' : (74, WSFloat.from_2, {'signed': False, 'scale': 0.1})},
            'wind_ave'      : {'bft': (76, WSInt.from_1, {'signed': False}),
                               'ms' : (77, WSFloat.from_1, {'signed': False, 'scale': 0.1})},
            'wind_gust'     : {'bft': (79, WSInt.from_1, {'signed': False}),
                               'ms' : (80, WSFloat.from_1, {'signed': False, 'scale': 0.1})},
            'wind_dir'      : (82, WSInt.wind_dir, {}),
            'rain'          : {'hour' : (83, WSFloat.from_2, {'signed': False, 'scale': 0.3}),
                               'day'  : (85, WSFloat.from_2, {'signed': False, 'scale': 0.3})},
            'time'          : (87, WSTime.from_raw, {}),
            'illuminance'   : (89, WSFloat.from_3, {'signed': False, 'scale': 0.1}),
            'uv'            : (92, WSInt.from_1, {'signed': False}),
            },
        'max'           : {
            'uv'            : {'val' : (93, WSInt.from_1, {'signed': False}),
                               'date': (6, WSDateTime.from_raw, {})},
            'illuminance'   : {'val' : (94, WSFloat.from_3, {'signed': False, 'scale': 0.1}),
                               'date': (11, WSDateTime.from_raw, {})},
            'hum_in'        : {'val' : (98, WSInt.from_1, {'signed': False}),
                               'date': (141, WSDateTime.from_raw, {})},
            'hum_out'       : {'val' : (100, WSInt.from_1, {'signed': False}),
                               'date': (151, WSDateTime.from_raw, {})},
            'temp_in'       : {'val' : (102, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (161, WSDateTime.from_raw, {})},
            'temp_out'      : {'val' : (106, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (171, WSDateTime.from_raw, {})},
            'windchill'     : {'val' : (110, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (181, WSDateTime.from_raw, {})},
            'dewpoint'      : {'val' : (114, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (191, WSDateTime.from_raw, {})},
            'abs_pressure'  : {'val' : (118, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'date': (201, WSDateTime.from_raw, {})},
            'rel_pressure'  : {'val' : (122, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'date': (211, WSDateTime.from_raw, {})},
            'wind_ave'      : {'val' : (126, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'date': (221, WSDateTime.from_raw, {})},
            'wind_gust'     : {'val' : (128, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'date': (226, WSDateTime.from_raw, {})},
            'rain'          : {
                'hour'          : {'val' : (130, WSFloat.from_2, {'signed': False, 'scale': 0.3}),
                                   'date': (231, WSDateTime.from_raw, {})},
                'day'           : {'val' : (132, WSFloat.from_2, {'signed': False, 'scale': 0.3}),
                                   'date': (236, WSDateTime.from_raw, {})},
                'week'          : {'val' : (134, WSFloat.from_2, {'signed': False, 'scale': 0.3}),
                                   'date': (241, WSDateTime.from_raw, {})},
                'month'         : {'val' : (136, WSFloat.from_2, {'signed': False, 'scale': 0.3, 'nibble_pos':140, 'nibble_high':True}),
                                   'date': (246, WSDateTime.from_raw, {})},
                'total'         : {'val' : (138, WSFloat.from_2, {'signed': False, 'scale': 0.3, 'nibble_pos':140, 'nibble_high':False}),
                                   'date': (251, WSDateTime.from_raw, {})},
                },
            },
        'min'           : {
            'hum_in'        : {'val' : (99, WSInt.from_1, {'signed': False}),
                               'date': (146, WSDateTime.from_raw, {})},
            'hum_out'       : {'val' : (101, WSInt.from_1, {'signed': False}),
                               'date': (156, WSDateTime.from_raw, {})},
            'temp_in'       : {'val' : (104, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (166, WSDateTime.from_raw, {})},
            'temp_out'      : {'val' : (108, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (176, WSDateTime.from_raw, {})},
            'windchill'     : {'val' : (112, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (186, WSDateTime.from_raw, {})},
            'dewpoint'      : {'val' : (116, WSFloat.from_2, {'signed': True, 'scale': 0.1}),
                               'date': (196, WSDateTime.from_raw, {})},
            'abs_pressure'  : {'val' : (120, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'date': (206, WSDateTime.from_raw, {})},
            'rel_pressure'  : {'val' : (124, WSFloat.from_2, {'signed': False, 'scale': 0.1}),
                               'date': (216, WSDateTime.from_raw, {})},
            },
        }
    fixed_format.update(lo_fix_format)
    # start of readings / end of fixed block
    data_start = 0x0100     # 256
    # bytes per reading, depends on weather station type
    reading_len = {
        '1080'  : 16,
        '3080'  : 20,
        }



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload weather data to PWS Weather.

`PWS Weather`_ is a site run by AerisWeather_ that "brings together
personal weather station data worldwide from locales not served by
primary weather services."

* Create account: http://www.pwsweather.com/register.php
* API based on WU protocol: `<http://wiki.wunderground.com/index.php/PWS_-_Upload_Protocol>`_
* Additional dependency: http://docs.python-requests.org/
* Example ``weather.ini`` configuration::

    [pwsweather]
    station = ABCDEFGH1
    password = xxxxxxx

    [logged]
    services = ['pwsweather', 'underground']

.. _PWS Weather: http://www.pwsweather.com/
.. _AerisWeather: https://www.aerisweather.com/

"""

from __future__ import absolute_import, unicode_literals

from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import sys

import requests

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.CatchupDataService):
    config = {
        'station' : ('', True, 'ID'),
        'password': ('', True, 'PASSWORD'),
        }
    fixed_data = {'action': 'updateraw', 'softwaretype': 'pywws'}
    logger = logger
    service_name = service_name
    template = """
#live#
#idx          "'dateutc'     : '%Y-%m-%d %H:%M:%S',"#
#wind_dir     "'winddir'     : '%.0f'," "" "winddir_degrees(x)"#
#wind_ave     "'windspeedmph': '%.2f'," "" "wind_mph(x)"#
#wind_gust    "'windgustmph' : '%.2f'," "" "wind_mph(x)"#
#hum_out      "'humidity'    : '%.d',"#
#temp_out     "'tempf'       : '%.1f'," "" "temp_f(x)"#
#rel_pressure "'baromin'     : '%.4f'," "" "pressure_inhg(x)"#
#calc "temp_f(dew_point(data['temp_out'], data['hum_out']))" "'dewptf': '%.1f',"#
#calc "rain_inch(rain_hour(data))" "'rainin': '%g',"#
#calc "rain_inch(rain_day(data))" "'dailyrainin': '%g',"#
"""

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # extend template
        if context.params.get('config', 'ws type') == '3080':
            self.template += """
#illuminance  "'solarradiation': '%.2f'," "" "illuminance_wm2(x)"#
#uv           "'UV'            : '%d',"#
"""

    @contextmanager
    def session(self):
        with requests.Session() as session:
            yield session

    def upload_data(self, session, prepared_data={}):
        try:
            rsp = session.get(
                'http://www.pwsweather.com/pwsupdate/pwsupdate.php',
                params=prepared_data, timeout=60)
        except Exception as ex:
            return False, repr(ex)
        if rsp.status_code != 200:
            return False, 'http status: {:d}'.format(rsp.status_code)
        text = rsp.text.strip()
        if text:
            return True, 'server response "{:s}"'.format(text)
        return True, 'OK'


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



"""
This module transfers and converts the pywws weather station data from one
data format and directory to another.

Usage:
python -m pywws.datastoretransfer SOURCE_TYPE SOURCE_DIR SINK_TYPE SINK_DIR

Example:
python -m pywws.datastoretransfer filedata c:\weather_data sqlite3data d:\weather

This can be used to convert from the default file base storage system to an
SQL based sorage system, or back. The tranfer will overwrite existing data
in place which may leave existing data in the destination if the incoming data
does not overlap (i.e. source data is newer than the destination). This is a
risky way to merge datastores together. Otherwise, its recommended to use the
optional -c argument to ensure the destination is cleared first.
You may choose the same storage module for both source and destination
with different directories, and this is the equivelent of simply copying the
data but will build the underlying files from scratch. However, copying the
files by hand is likely to be faster.

Config files such as weather.ini, and folders such as templates etc.
are not copied, so please ensure you consider seperately if required.
"""

from __future__ import absolute_import, print_function

import argparse
import importlib
import logging

import pywws.logger

logger = logging.getLogger(__name__)
pywws.logger.setup_handler(1)

def monitor(i):
    """Given an iterator, yields data from it
    but prints progress every 10,000 records"""
    count = 0
    for x in i:
        count+=1
        if count % 10000 == 0:
            logger.info("%d records so far, current record is %s",
                count, x["idx"])
        yield x

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""Copy pywws data from one storage system to another.
        You must specify the pywws data module, and the path for both
        the source data set, and the destination data set.""")
    parser.add_argument("-c",
        help="Clear the existing destination datastore before transfer",
        action="store_true", dest="clearfirst")
    parser.add_argument("SourceType",
        help="The source storage system type")
    parser.add_argument("SourceDir",
        help="The source directory")
    parser.add_argument("SinkType",
        help="The destination storage system type")
    parser.add_argument("SinkDir",
        help="The destination directory")

    args = parser.parse_args()
    source_type = args.SourceType
    sink_type = args.SinkType
    source_dir = args.SourceDir
    sink_dir = args.SinkDir
    clearfirst=args.clearfirst

    if source_type == sink_type and source_dir == sink_dir:
        raise ValueError("You have specified the same source and sink")

    Source = importlib.import_module("."+source_type, package="pywws")
    Sink = importlib.import_module("."+sink_type, package="pywws")

    RawSink = Sink.RawStore(sink_dir)
    if clearfirst:
        logger.info("Clearing destination Raw Data...")
        RawSink.clear()
    logger.info("Transfering Raw Data...")
    RawSink.update(monitor(Source.RawStore(source_dir)))
    RawSink.flush()

    CalibSink = Sink.CalibStore(sink_dir)
    if clearfirst:
        logger.info("Clearing destination Calibrated Data...")
        CalibSink.clear()
    logger.info("Transfering Calibrated Data...")
    CalibSink.update(monitor(Source.CalibStore(source_dir)))
    CalibSink.flush()

    HourlySink = Sink.HourlyStore(sink_dir)
    if clearfirst:
        logger.info("Clearing destination Hourly Data...")
        HourlySink.clear()
    logger.info("Transfering Hourly Data...")
    HourlySink.update(monitor(Source.HourlyStore(source_dir)))
    HourlySink.flush()

    DailySink = Sink.DailyStore(sink_dir)
    if clearfirst:
        logger.info("Clearing destination Daily Data...")
        DailySink.clear()
    logger.info("Transfering Daily Data...")
    DailySink.update(monitor(Source.DailyStore(source_dir)))
    DailySink.flush()

    MonthlySink = Sink.MonthlyStore(sink_dir)
    if clearfirst:
        logger.info("Clearing destination Monthly Data...")
        MonthlySink.clear()
    logger.info("Transfering Monthly Data...")
    MonthlySink.update(monitor(Source.MonthlyStore(source_dir)))
    MonthlySink.flush()

    logger.info("Done!")